package net.sergeych.ling

data class Pos(val source: Source, val line: Int, val column: Int) {
    override fun toString(): String {
        return "${source.fileName}:$line:$column"
    }

    fun back(): Pos =
        if( column > 0) Pos(source, line, column-1)
        else if( line > 0) Pos(source, line-1, source.lines[line-1].length - 1)
        else throw IllegalStateException("can't go back from line 0, column 0")

    val currentLine: String get() = source.lines[line]

    val showSource: String by lazy {
        source.lines[line] + "\n" + "-".repeat(column - 1) + "^\n"
    }

    companion object {
        val builtIn = Pos(Source.builtIn, 0, 0)
    }

}

class MutablePos(private val from: Pos) {

    val lines = from.source.lines
    var line = from.line
        private set
    var column = from.column
        private set

    val end: Boolean get() =
        line == lines.size

    fun reset(to: Pos) {
        line = to.line
        column = to.column
    }

    fun toPos(): Pos = Pos(from.source, line, column)

    fun advance(): Char? {
        if (end) return null
        val current = lines[line]
        return if (column+1 < current.length) {
            current[column++]
        } else {
            line++
            column = 0
            if(line >= lines.size) null
            else lines[line][column]
        }
    }

    fun back() {
        if (column > 0) column--
        else if (line > 0)
            column = lines[--line].length - 1
        else throw IllegalStateException("can't go back from line 0, column 0")
    }
    val currentChar: Char
        get() =
            if (end) 0.toChar()
            else lines[line][column]

    override fun toString() = "($line:$column)"

}
